/*
 * Copyright 2007 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.xbib.files.ftp.mock.stub.command;

import java.util.Objects;
import java.util.logging.Level;
import org.xbib.files.ftp.mock.core.MockFtpServerException;
import org.xbib.files.ftp.mock.core.command.Command;
import org.xbib.files.ftp.mock.core.command.CommandHandler;
import org.xbib.files.ftp.mock.core.command.InvocationRecord;
import org.xbib.files.ftp.mock.core.session.Session;
import org.xbib.files.ftp.mock.core.util.Assert;
import org.xbib.files.ftp.mock.core.util.AssertFailedException;

import java.io.IOException;
import java.io.InputStream;
import java.util.logging.Logger;

/**
 * CommandHandler for the RETR command. Returns the contents of the specified file on the
 * data connection, along with two replies on the control connection: a reply code of 150 and
 * another of 226.
 *
 * <p>The <code>file</code> property specifies the pathname for the file whose contents should
 * be returned from this command. The file path is relative to the CLASSPATH (using the
 * ClassLoader for this class).
 *
 * <p>An exception is thrown if the <code>file</code> property has not been set or if the specified
 * file does not exist or cannot be read.
 *
 * <p>Each invocation record stored by this CommandHandler includes the following data element key/values:
 * <ul>
 * <li>{@link #PATHNAME_KEY} ("pathname") - the pathname of the file submitted on the invocation (the first command parameter)
 * </ul>
 *
 * @author Chris Mair
 */
public class FileRetrCommandHandler extends AbstractStubDataCommandHandler implements CommandHandler {

    private static final Logger logger = Logger.getLogger(FileRetrCommandHandler.class.getName());

    public static final String PATHNAME_KEY = "pathname";
    public static final int BUFFER_SIZE = 512;     // package-private for testing

    private final Class<?> cl;

    private String file;

    /**
     * Create new uninitialized instance
     */
    public FileRetrCommandHandler() {
        this(null, FileRetrCommandHandler.class);
    }

    /**
     * Create new instance using the specified file pathname
     *
     * @param file - the path to the file
     * @throws AssertFailedException - if the file is null
     */
    public FileRetrCommandHandler(String file, Class<?> cl) {
        this.file = file;
        this.cl = Objects.requireNonNull(cl);
    }

    /**
     * @see AbstractStubDataCommandHandler#beforeProcessData(Command, Session, InvocationRecord)
     */
    protected void beforeProcessData(Command command, Session session, InvocationRecord invocationRecord) throws Exception {
        Assert.notNull(file, "file");
        invocationRecord.set(PATHNAME_KEY, command.getRequiredParameter(0));
    }

    /**
     * @see AbstractStubDataCommandHandler#processData(Command, Session, InvocationRecord)
     */
    protected void processData(Command command, Session session, InvocationRecord invocationRecord) {
        logger.log(Level.INFO, "cl = " + cl + " file = " + file);
        InputStream inputStream = cl.getClassLoader().getResourceAsStream(file);
        Assert.notNull(inputStream, "InputStream for [" + file + "]");
        byte[] buffer = new byte[BUFFER_SIZE];
        try {
            int numBytes;
            while ((numBytes = inputStream.read(buffer)) != -1) {
                LOG.log(Level.FINEST, "Sending " + numBytes + " bytes...");
                session.sendData(buffer, numBytes);
            }
        }
        catch (IOException e) {
            throw new MockFtpServerException(e);
        }
    }

    /**
     * Set the path of the file whose contents should be returned when this command is
     * invoked. The path is relative to the CLASSPATH.
     *
     * @param file - the path to the file
     * @throws AssertFailedException - if the file is null
     */
    public void setFile(String file) {
        this.file = file;
    }

}
