package org.apache.sshd.common.util.security.edec;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.math.BigInteger;
import java.security.GeneralSecurityException;
import java.security.KeyFactory;
import java.security.KeyPairGenerator;
import java.security.interfaces.EdECPrivateKey;
import java.security.interfaces.EdECPublicKey;
import java.security.spec.EdECPoint;
import java.security.spec.EdECPrivateKeySpec;
import java.security.spec.EdECPublicKeySpec;
import java.security.spec.NamedParameterSpec;
import java.util.Collections;
import java.util.Map;
import org.apache.sshd.common.config.keys.KeyEntryResolver;
import org.apache.sshd.common.config.keys.impl.AbstractPublicKeyEntryDecoder;
import org.apache.sshd.common.keyprovider.KeyPairProvider;
import org.apache.sshd.common.session.SessionContext;

public final class Ed25519PublicKeyDecoder extends AbstractPublicKeyEntryDecoder<EdECPublicKey, EdECPrivateKey> {

    public static final int MAX_ALLOWED_SEED_LEN = 1024;

    public static final Ed25519PublicKeyDecoder INSTANCE = new Ed25519PublicKeyDecoder();

    private Ed25519PublicKeyDecoder() {
        super(EdECPublicKey.class, EdECPrivateKey.class, Collections.singletonList(KeyPairProvider.SSH_ED25519));
    }

    @Override
    public EdECPublicKey clonePublicKey(EdECPublicKey key) throws GeneralSecurityException {
        if (key == null) {
            return null;
        } else {
            NamedParameterSpec namedParameterSpec = NamedParameterSpec.ED25519;
            EdECPoint edECPoint = key.getPoint();
            return generatePublicKey(new EdECPublicKeySpec(namedParameterSpec, edECPoint));
        }
    }

    @Override
    public EdECPrivateKey clonePrivateKey(EdECPrivateKey key) throws GeneralSecurityException {
        if (key == null) {
            return null;
        } else {
            NamedParameterSpec namedParameterSpec = NamedParameterSpec.ED25519;
            return generatePrivateKey(new EdECPrivateKeySpec(namedParameterSpec, key.getEncoded()));
        }
    }

    @Override
    public KeyPairGenerator getKeyPairGenerator() throws GeneralSecurityException {
        return KeyPairGenerator.getInstance("Ed25519");
    }

    @Override
    public String encodePublicKey(OutputStream s, EdECPublicKey key) throws IOException {
        // TODO
        return "ssh-ed25519";
    }

    @Override
    public KeyFactory getKeyFactoryInstance() throws GeneralSecurityException {
        return KeyFactory.getInstance("EdDSA");
    }

    @Override
    public EdECPublicKey decodePublicKey(SessionContext session,
                                         String keyType,
                                         InputStream keyData,
                                         Map<String, String> headers)
            throws IOException, GeneralSecurityException {
        byte[] pk = KeyEntryResolver.readRLEBytes(keyData, MAX_ALLOWED_SEED_LEN);
        boolean xisodd = false;
        int lastbyteInt = pk[pk.length - 1];
        if ((lastbyteInt & 255) >> 7 == 1) {
            xisodd = true;
        }
        pk[pk.length - 1] &= 127;
        BigInteger y = new BigInteger(1, pk);
        NamedParameterSpec namedParameterSpec = NamedParameterSpec.ED25519;
        EdECPoint ep = new EdECPoint(xisodd, y);
        EdECPublicKeySpec spec = new EdECPublicKeySpec(namedParameterSpec, ep);
        return EdECPublicKey.class.cast(generatePublicKey(spec));
    }
}
