package org.apache.sshd.common.util.security.edec;

import java.io.IOException;
import java.io.InputStream;
import java.io.StreamCorruptedException;
import java.math.BigInteger;
import java.security.GeneralSecurityException;
import java.security.InvalidKeyException;
import java.security.KeyFactory;
import java.security.KeyPairGenerator;
import java.security.interfaces.EdECPrivateKey;
import java.security.interfaces.EdECPublicKey;
import java.security.spec.EdECPoint;
import java.security.spec.EdECPrivateKeySpec;
import java.security.spec.EdECPublicKeySpec;
import java.security.spec.NamedParameterSpec;
import java.util.Arrays;
import java.util.Collections;
import java.util.Locale;
import java.util.Objects;
import org.apache.sshd.common.config.keys.FilePasswordProvider;
import org.apache.sshd.common.config.keys.KeyEntryResolver;
import org.apache.sshd.common.config.keys.impl.AbstractPrivateKeyEntryDecoder;
import org.apache.sshd.common.keyprovider.KeyPairProvider;
import org.apache.sshd.common.session.SessionContext;
import org.apache.sshd.common.util.GenericUtils;
import org.apache.sshd.common.util.io.SecureByteArrayOutputStream;
import org.apache.sshd.common.util.io.der.ASN1Object;
import org.apache.sshd.common.util.io.der.ASN1Type;
import org.apache.sshd.common.util.io.der.DERParser;


public class OpenSSHEd25519PrivateKeyEntryDecoder extends AbstractPrivateKeyEntryDecoder<EdECPublicKey, EdECPrivateKey> {

    public static final OpenSSHEd25519PrivateKeyEntryDecoder INSTANCE = new OpenSSHEd25519PrivateKeyEntryDecoder();

    private static final int PK_SIZE = 32;

    private static final int SK_SIZE = 32;

    private static final int KEYPAIR_SIZE = PK_SIZE + SK_SIZE;

    public OpenSSHEd25519PrivateKeyEntryDecoder() {
        super(EdECPublicKey.class, EdECPrivateKey.class, Collections.singletonList(KeyPairProvider.SSH_ED25519));
    }

    @Override
    public EdECPrivateKey decodePrivateKey(SessionContext session,
                                           String keyType,
                                           FilePasswordProvider passwordProvider,
                                           InputStream keyData)
            throws IOException, GeneralSecurityException {
        if (!"ssh-ed25519".equals(keyType)) {
            throw new InvalidKeyException("Unsupported key type: " + keyType);
        }
        // ed25519 bernstein naming: pk .. public key, sk .. secret key
        // we expect to find two byte arrays with the following structure (type:size):
        // [pk:32], [sk:32,pk:32]
        byte[] pk = GenericUtils.EMPTY_BYTE_ARRAY;
        byte[] keypair = GenericUtils.EMPTY_BYTE_ARRAY;
        try {
            pk = KeyEntryResolver.readRLEBytes(keyData, PK_SIZE * 2);
            keypair = KeyEntryResolver.readRLEBytes(keyData, KEYPAIR_SIZE * 2);
            if (pk.length != PK_SIZE) {
                throw new InvalidKeyException(
                        String.format(Locale.ENGLISH, "Unexpected pk size: %s (expected %s)", pk.length, PK_SIZE));
            }
            if (keypair.length != KEYPAIR_SIZE) {
                throw new InvalidKeyException(
                        String.format(Locale.ENGLISH, "Unexpected keypair size: %s (expected %s)", keypair.length,
                                KEYPAIR_SIZE));
            }
            // verify that the keypair contains the expected pk
            // yes, it's stored redundant, this seems to mimic the output structure of the keypair generation interface
            if (!Arrays.equals(pk, Arrays.copyOfRange(keypair, SK_SIZE, KEYPAIR_SIZE))) {
                throw new InvalidKeyException("Keypair did not contain the public key.");
            }
            byte[] seed = Arrays.copyOf(keypair, SK_SIZE);
            NamedParameterSpec spec = NamedParameterSpec.ED25519;
            EdECPrivateKeySpec keySpec = new EdECPrivateKeySpec(spec, seed);
            KeyFactory factory = KeyFactory.getInstance("EdDSA");
            return EdECPrivateKey.class.cast(factory.generatePrivate(keySpec));
        } finally {
            // get rid of sensitive data a.s.a.p
            Arrays.fill(pk, (byte) 0);
            Arrays.fill(keypair, (byte) 0);
        }
    }

    @Override
    public String encodePrivateKey(SecureByteArrayOutputStream s,
                                   EdECPrivateKey key,
                                   EdECPublicKey pubKey) throws IOException {
        Objects.requireNonNull(key, "No private key provided");
        return "ssh-ed25519";
    }

    @Override
    public boolean isPublicKeyRecoverySupported() {
        return true;
    }

    @Override
    public EdECPublicKey recoverPublicKey(EdECPrivateKey key) throws GeneralSecurityException {
        byte[] pk = key.getEncoded();
        boolean xisodd = false;
        int lastbyteInt = pk[pk.length - 1];
        if ((lastbyteInt & 255) >> 7 == 1) {
            xisodd = true;
        }
        pk[pk.length - 1] &= 127;
        BigInteger y = new BigInteger(1, pk);
        NamedParameterSpec paramSpec = new NamedParameterSpec("Ed25519");
        EdECPoint ep = new EdECPoint(xisodd, y);
        EdECPublicKeySpec publicKeySpec = new EdECPublicKeySpec(paramSpec, ep);
        KeyFactory factory = KeyFactory.getInstance("EdDSA");
        return EdECPublicKey.class.cast(factory.generatePublic(publicKeySpec));

    }

    @Override
    public EdECPublicKey clonePublicKey(EdECPublicKey key) throws GeneralSecurityException {
        if (key == null) {
            return null;
        } else {
            NamedParameterSpec namedParameterSpec = NamedParameterSpec.ED25519;
            EdECPoint edECPoint = key.getPoint();
            EdECPublicKeySpec spec = new EdECPublicKeySpec(namedParameterSpec, edECPoint);
            return generatePublicKey(spec);
        }
    }

    @Override
    public EdECPrivateKey clonePrivateKey(EdECPrivateKey key) throws GeneralSecurityException {
        if (key == null) {
            return null;
        } else {
            return getPrivateKey(key.getEncoded());
        }
    }

    @Override
    public KeyPairGenerator getKeyPairGenerator() throws GeneralSecurityException {
        return KeyPairGenerator.getInstance("Ed25519");
    }

    @Override
    public KeyFactory getKeyFactoryInstance() throws GeneralSecurityException {
        return KeyFactory.getInstance("EdDSA");
    }

    private static EdECPrivateKey getPrivateKey(byte[] keyData) throws GeneralSecurityException {
        try (DERParser parser = new DERParser(keyData)) {
            ASN1Object obj = parser.readObject();
            if (obj == null) {
                throw new StreamCorruptedException("Missing key data container");
            }
            ASN1Type objType = obj.getObjType();
            if (objType != ASN1Type.OCTET_STRING) {
                throw new StreamCorruptedException("Mismatched key data container type: " + objType);
            }
            return generatePrivateKey(obj.getValue());
        } catch (IOException e) {
            throw new GeneralSecurityException(e);
        }
    }

    private static EdECPrivateKey generatePrivateKey(byte[] seed) throws GeneralSecurityException {
        NamedParameterSpec spec = NamedParameterSpec.ED25519;
        EdECPrivateKeySpec keySpec = new EdECPrivateKeySpec(spec, seed);
        KeyFactory factory = KeyFactory.getInstance("EdDSA");
        return EdECPrivateKey.class.cast(factory.generatePrivate(keySpec));
    }

}
